// Copyright (c) 2014 Cognivue Corporation
// All rights reserved.
// 
// WARRANTY DISCLAIMER
// 
// THESE MATERIALS ARE PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
// OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THESE
// MATERIALS, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
// Cognivue Corporation is the author of the Materials, and requests that all
// problem reports or change requests be submitted to it directly

#include "Bitonic_APU.h"

// stage is equal to 0
void func0(__local int4 *pSrc0, const size_t g_id, const uint dir)
{
	int4 imask0 = (int4)(0, -1, -1,  0);
	int4 imask10 = (int4)(0,  0, -1, -1);
	int4 imask11 = (int4)(0, -1,  0, -1);
	int4 srcLeft, srcRight, mask;
	//const int i = get_local_id(0);

	srcLeft = pSrc0[0];
	srcRight = srcLeft.yxwz;
	mask = (srcLeft < srcRight) ^ imask0;
	if( dir )
		srcLeft = (srcLeft & mask) | (srcRight & ~mask);
	else
		srcLeft = (srcLeft & ~mask) | (srcRight & mask);

	srcRight = srcLeft.zwxy;
	mask = (srcLeft < srcRight) ^ imask10;

	if((g_id & 1) ^ dir)
	{
		srcLeft = (srcLeft & mask) | (srcRight & ~mask);
		srcRight = srcLeft.yxwz;
		mask = (srcLeft < srcRight) ^ imask11;
		pSrc0[0] = (srcLeft & mask) | (srcRight & ~mask);
	}
	else
	{
		srcLeft = (srcLeft & ~mask) | (srcRight & mask);
		srcRight = srcLeft.yxwz;
		mask = (srcLeft < srcRight) ^ imask11;
		pSrc0[0] = (srcLeft & ~mask) | (srcRight & mask);
	}
}

// stage is not equal to 0, passOfStage is equal to 0
void func1(__local int4 *pSrc0, const size_t g_id, const uint stage, const uint dir)
{

	int4 imask10 = (int4)(0,  0, -1, -1);
	int4 imask11 = (int4)(0, -1,  0, -1);
	int4 srcLeft, srcRight, mask;

	srcLeft = pSrc0[0];
	srcRight = srcLeft.zwxy;
	mask = (srcLeft < srcRight) ^ imask10;

	if(((g_id >> stage) & 1) ^ dir)
	{
		srcLeft = (srcLeft & mask) | (srcRight & ~mask);
		srcRight = srcLeft.yxwz;
		mask = (srcLeft < srcRight) ^ imask11;
		pSrc0[0] = (srcLeft & mask) | (srcRight & ~mask);
	}
	else
	{
		srcLeft = (srcLeft & ~mask) | (srcRight & mask);
		srcRight = srcLeft.yxwz;
		mask = (srcLeft < srcRight) ^ imask11;
		pSrc0[0] = (srcLeft & ~mask) | (srcRight & mask);
	}
}

// stage is not equal to 0, passOfStage is larger than 0
void func2(__local int4 *pSrcLeft, 
               __local int4 *pSrcRight, 
			   //const size_t l_id,
			   const size_t g_id,
			   //const uint passOfStage, 
			   const uint stage, 
			   const uint dir)
{
	int4 srcLeft, srcRight, mask;

	srcLeft = pSrcLeft[0];
	srcRight = pSrcRight[0];
	mask = srcLeft < srcRight;

    //barrier(CLK_LOCAL_MEM_FENCE); 			

	int4 imin = (srcLeft & mask) | (srcRight & ~mask);
	int4 imax = (srcLeft & ~mask) | (srcRight & mask);
			
    //barrier(CLK_LOCAL_MEM_FENCE); 

	if( ((g_id>>(stage-1)) & 1) ^ dir )
	{
		pSrcLeft[0]  = imin;
		pSrcRight[0] = imax;
	}
	else
	{
		pSrcRight[0] = imin;
		pSrcLeft[0]  = imax;
	}
}

__kernel void __attribute__((vec_type_hint(int4))) 
              __attribute__((reqd_work_group_size(BITONIC_LOCAL_SIZE, 1, 1)))
              BitonicSort_WG(__global int4 * theArray,
						     const uint stage,
						     const uint passOfStage,
						     const uint dir)
{
	// global id
	size_t i = get_global_id(0);
	int4 srcLeft, srcRight, mask;
	int4 imask10 = (int4)(0,  0, -1, -1);
	int4 imask11 = (int4)(0, -1,  0, -1);

	// local id 
	const int lx = get_local_id(0);

	// work group id
	const int wx = get_group_id(0);

	// local size
	const int local_size = get_local_size(0);

	__local int4 input0_local[BITONIC_LOCAL_SIZE*2]; 

	__local int4 *pSrc0;
	__local int4 *pSrc1;
	__local int4 *pLeft;
	__local int4 *pRight;

	uint src_offset0, src_offset1;
	uint n_thPowerOf2=0, temp;

	event_t ievent;
	event_t oevent;
#if 0
	src_offset0 = wx * local_size;
	pSrc0 = input0_local;

	ievent = async_work_group_copy(pSrc0, theArray+src_offset0, local_size, 0);
	wait_group_events(1, &ievent);

	//oevent = async_work_group_copy(theArray+src_offset0, pSrc0, local_size, 0);
	//wait_group_events(1, &oevent);
#else
	if(stage > 0)
	{
		if(passOfStage > 0)	//upper level pass, exchange between two fours
		{
			temp = local_size;

			while(temp)
			{
				temp >>= 1;
				n_thPowerOf2++;
			};

			if(passOfStage < n_thPowerOf2)  // If local size is changed, it also should be changed!!!
			{
				// 1 source & 1 DMA
				src_offset0 = 2* wx * local_size;

				pSrc0 = input0_local;

                //barrier(CLK_LOCAL_MEM_FENCE);
				//barrier(CLK_GLOBAL_MEM_FENCE);
				// Saving global data to local memory 
				ievent = async_work_group_copy(pSrc0, theArray+src_offset0, local_size*2, 0);
				wait_group_events(1, &ievent);

                size_t r = 1 << (passOfStage - 1);
                size_t lmask = r - 1;
				size_t left = ((lx>>(passOfStage-1)) << passOfStage) + (lx & lmask);
				size_t right = left + r;
				//int4 srcLeft, srcRight, mask;

				pLeft = &(input0_local[left]);
				pRight = &(input0_local[right]);

				//func2(pLeft, pRight, lx, i, passOfStage, stage, dir);
				//func2(pLeft, pRight, i, passOfStage, stage, dir);
				func2(pLeft, pRight, i, stage, dir);

				//barrier(CLK_LOCAL_MEM_FENCE);

				oevent = async_work_group_copy(theArray+src_offset0, pSrc0, local_size*2, 0);
				wait_group_events(1, &oevent);
			}
			else // passOfStage >= 8
			{
				// 2 sources & 2 DMAs
				uint d = (uint)passOfStage-n_thPowerOf2;  // If local size is changed, it also should be changed!!!

				src_offset0 = (i-lx)+ ((wx>>d)<<d)*local_size;
				src_offset1 = (i-lx)+ ((wx>>d)<<d)*local_size + (1<<d)*local_size;

				pSrc0 = input0_local;
				pSrc1 = input0_local+local_size;

				//barrier(CLK_LOCAL_MEM_FENCE); 

				ievent = async_work_group_copy(pSrc0, theArray+src_offset0, local_size, 0);
				wait_group_events(1, &ievent);	

				ievent = async_work_group_copy(pSrc1, theArray+src_offset1, local_size, 0);
				wait_group_events(1, &ievent);

				pLeft = &(pSrc0[lx]);
				pRight = &(pSrc1[lx]);

				//func2(pLeft, pRight, lx, i, passOfStage, stage, dir);
				//func2(pLeft, pRight, i, passOfStage, stage, dir);
				func2(pLeft, pRight, i, stage, dir);

				//barrier(CLK_GLOBAL_MEM_FENCE);
				//barrier(CLK_LOCAL_MEM_FENCE); 

				oevent = async_work_group_copy(theArray+src_offset0, pSrc0, local_size, 0);
				wait_group_events(1, &oevent);	

				oevent = async_work_group_copy(theArray+src_offset1, pSrc1, local_size, 0);
				wait_group_events(1, &oevent);
			}
		}
		else	//last pass, sort inside one four
		{
			// 1 source & 1 DMA
			src_offset0 = wx * local_size;

			pSrc0 = input0_local;
			//barrier(CLK_LOCAL_MEM_FENCE);

			// Saving global data to local memory 
			ievent = async_work_group_copy(pSrc0, theArray+src_offset0, local_size, 0);
			wait_group_events(1, &ievent);

			func1(&(pSrc0[lx]), i, stage, dir);

			//barrier(CLK_GLOBAL_MEM_FENCE);
			//barrier(CLK_LOCAL_MEM_FENCE);
			// Saving local data to global memory 
			oevent = async_work_group_copy(theArray+src_offset0, pSrc0, local_size, 0);
			wait_group_events(1, &oevent);

		}
	}
	else	//first stage, sort inside one four
	{
		// 1 source & 1 DMA
		src_offset0 = wx * local_size;

		pSrc0 = input0_local;
		//barrier(CLK_LOCAL_MEM_FENCE);
		// Saving global data to local memory 
		ievent = async_work_group_copy(pSrc0, theArray+src_offset0, local_size, 0);
		wait_group_events(1, &ievent);

		func0(&(pSrc0[lx]), i, dir);

		//barrier(CLK_GLOBAL_MEM_FENCE);
		//barrier(CLK_LOCAL_MEM_FENCE);
		// Saving local data to global memory 
		oevent = async_work_group_copy(theArray+src_offset0, pSrc0, local_size, 0);
		wait_group_events(1, &oevent);
	}
#endif
}

void func00(__local int4 *pSrc0_in, __local int4 *pDst0_out, const size_t g_id, const uint dir)
{
	int4 imask0 = (int4)(0, -1, -1,  0);
	int4 imask10 = (int4)(0,  0, -1, -1);
	int4 imask11 = (int4)(0, -1,  0, -1);
	int4 srcLeft, srcRight, mask;
	//const int i = get_local_id(0);

	srcLeft = pSrc0_in[0];
	srcRight = srcLeft.yxwz;
	mask = (srcLeft < srcRight) ^ imask0;
	if( dir )
		srcLeft = (srcLeft & mask) | (srcRight & ~mask);
	else
		srcLeft = (srcLeft & ~mask) | (srcRight & mask);

	srcRight = srcLeft.zwxy;
	mask = (srcLeft < srcRight) ^ imask10;

	if((g_id & 1) ^ dir)
	{
		srcLeft = (srcLeft & mask) | (srcRight & ~mask);
		srcRight = srcLeft.yxwz;
		mask = (srcLeft < srcRight) ^ imask11;
		pDst0_out[0] = (srcLeft & mask) | (srcRight & ~mask);
	}
	else
	{
		srcLeft = (srcLeft & ~mask) | (srcRight & mask);
		srcRight = srcLeft.yxwz;
		mask = (srcLeft < srcRight) ^ imask11;
		pDst0_out[0] = (srcLeft & ~mask) | (srcRight & mask);
	}
}

// stage is not equal to 0, passOfStage is equal to 0
void func01(__local int4 *pSrc0_in, __local int4 *pDst0_out, const size_t g_id, const uint stage, const uint dir)
{

	int4 imask10 = (int4)(0,  0, -1, -1);
	int4 imask11 = (int4)(0, -1,  0, -1);
	int4 srcLeft, srcRight, mask;

	srcLeft = pSrc0_in[0];
	srcRight = srcLeft.zwxy;
	mask = (srcLeft < srcRight) ^ imask10;

	if(((g_id >> stage) & 1) ^ dir)
	{
		srcLeft = (srcLeft & mask) | (srcRight & ~mask);
		srcRight = srcLeft.yxwz;
		mask = (srcLeft < srcRight) ^ imask11;
		pDst0_out[0] = (srcLeft & mask) | (srcRight & ~mask);
	}
	else
	{
		srcLeft = (srcLeft & ~mask) | (srcRight & mask);
		srcRight = srcLeft.yxwz;
		mask = (srcLeft < srcRight) ^ imask11;
		pDst0_out[0] = (srcLeft & ~mask) | (srcRight & mask);
	}
}

// stage is not equal to 0, passOfStage is larger than 0
void func02(__local int4 *pSrcLeft_in, 
            __local int4 *pSrcRight_in,
			__local int4 *pSrcLeft_out, 
            __local int4 *pSrcRight_out, 
			//const size_t l_id,
			const size_t g_id,
			//const uint passOfStage, 
			const uint stage, 
			const uint dir)
{
	int4 srcLeft, srcRight, mask;

	srcLeft = pSrcLeft_in[0];
	srcRight = pSrcRight_in[0];
	mask = srcLeft < srcRight;

    //barrier(CLK_LOCAL_MEM_FENCE); 			

	int4 imin = (srcLeft & mask) | (srcRight & ~mask);
	int4 imax = (srcLeft & ~mask) | (srcRight & mask);
			
    //barrier(CLK_LOCAL_MEM_FENCE); 

	if( ((g_id>>(stage-1)) & 1) ^ dir )
	{
		pSrcLeft_out[0]  = imin;
		pSrcRight_out[0] = imax;
	}
	else
	{
		pSrcRight_out[0] = imin;
		pSrcLeft_out[0]  = imax;
	}
}

__kernel void __attribute__((vec_type_hint(int4))) 
              __attribute__((reqd_work_group_size(BITONIC_LOCAL_SIZE, 1, 1)))
              BitonicSort_WG_2(__global int4 * theArray,
						       const uint stage,
						       const uint passOfStage,
						       const uint dir)
{
	// global id
	size_t i = get_global_id(0);
	int4 srcLeft, srcRight, mask;
	int4 imask10 = (int4)(0,  0, -1, -1);
	int4 imask11 = (int4)(0, -1,  0, -1);

	// local id 
	const int lx = get_local_id(0);

	// work group id
	const int wx = get_group_id(0);

	// local size
	const int local_size = get_local_size(0);

	__local int4 input0_local[BITONIC_LOCAL_SIZE*2]; 
	__local int4 output0_local[BITONIC_LOCAL_SIZE*2];

	__local int4 *pSrc0_in;
	__local int4 *pSrc1_in;
	__local int4 *pLeft_in;
	__local int4 *pRight_in;

	__local int4 *pDst0_out;
	__local int4 *pDst1_out;
	__local int4 *pLeft_out;
	__local int4 *pRight_out;

	uint src_offset0, src_offset1;
	uint n_thPowerOf2=0, temp;

	event_t ievent;
	event_t oevent;

	if(stage > 0)
	{
		if(passOfStage > 0)	//upper level pass, exchange between two fours
		{
			temp = local_size;

			while(temp)
			{
				temp >>= 1;
				n_thPowerOf2++;
			};

			if(passOfStage < n_thPowerOf2)  // If local size is changed, it also should be changed!!!
			{
				// 1 source & 1 DMA
				src_offset0 = 2* wx * local_size;

				pSrc0_in = input0_local;
				pDst0_out = output0_local;

                //barrier(CLK_LOCAL_MEM_FENCE);
				//barrier(CLK_GLOBAL_MEM_FENCE);
				// Saving global data to local memory 
				ievent = async_work_group_copy(pSrc0_in, theArray+src_offset0, local_size*2, 0);
				wait_group_events(1, &ievent);

                size_t r = 1 << (passOfStage - 1);
                size_t lmask = r - 1;
				size_t left = ((lx>>(passOfStage-1)) << passOfStage) + (lx & lmask);
				size_t right = left + r;
				//int4 srcLeft, srcRight, mask;

				pLeft_in = &(input0_local[left]);
				pRight_in = &(input0_local[right]);

				pLeft_out = &(output0_local[left]);
				pRight_out = &(output0_local[right]);

				//func2(pLeft_in, pRight_in, i, stage, dir);
				//for(int i=0; i<local_size*2; i++)
				//	output0_local[i] = input0_local[i];
				func02(pLeft_in, pRight_in, pLeft_out, pRight_out, i, stage, dir);

				barrier(CLK_LOCAL_MEM_FENCE);

				oevent = async_work_group_copy(theArray+src_offset0, pDst0_out, local_size*2, 0);
				wait_group_events(1, &oevent);
			}
			else // passOfStage >= 8
			{
				// 2 sources & 2 DMAs
				uint d = (uint)passOfStage-n_thPowerOf2;  // If local size is changed, it also should be changed!!!

				src_offset0 = (i-lx)+ ((wx>>d)<<d)*local_size;
				src_offset1 = (i-lx)+ ((wx>>d)<<d)*local_size + (1<<d)*local_size;

				pSrc0_in = input0_local;
				pSrc1_in = input0_local+local_size;
				pDst0_out = output0_local;
				pDst1_out = output0_local+local_size;

				//barrier(CLK_LOCAL_MEM_FENCE); 

				ievent = async_work_group_copy(pSrc0_in, theArray+src_offset0, local_size, 0);
				wait_group_events(1, &ievent);	

				ievent = async_work_group_copy(pSrc1_in, theArray+src_offset1, local_size, 0);
				wait_group_events(1, &ievent);

				pLeft_in = &(pSrc0_in[lx]);
				pRight_in = &(pSrc1_in[lx]);
				pLeft_out = &(pDst0_out[lx]);
				pRight_out = &(pDst1_out[lx]);

				//func2(pLeft_in, pRight_in, i, stage, dir);
				//for(int i=0; i<local_size*2; i++)
				//	output0_local[i] = input0_local[i];

				func02(pLeft_in, pRight_in, pLeft_out, pRight_out, i, stage, dir);

				//barrier(CLK_GLOBAL_MEM_FENCE);
				barrier(CLK_LOCAL_MEM_FENCE); 

				oevent = async_work_group_copy(theArray+src_offset0, pDst0_out, local_size, 0);
				wait_group_events(1, &oevent);	

				oevent = async_work_group_copy(theArray+src_offset1, pDst1_out, local_size, 0);
				wait_group_events(1, &oevent);
			}
		}
		else	//last pass, sort inside one four
		{
			// 1 source & 1 DMA
			src_offset0 = wx * local_size;

			pSrc0_in = input0_local;
			pDst0_out = output0_local;
			//barrier(CLK_LOCAL_MEM_FENCE);

			// Saving global data to local memory 
			ievent = async_work_group_copy(pSrc0_in, theArray+src_offset0, local_size, 0);
			wait_group_events(1, &ievent);

			func01(&(pSrc0_in[lx]), &(pDst0_out[lx]), i, stage, dir);
			//func1(&(pSrc0_in[lx]), i, stage, dir);
			//for(int i=0; i<local_size*2; i++)
			//	output0_local[i] = input0_local[i];

			//barrier(CLK_GLOBAL_MEM_FENCE);
			barrier(CLK_LOCAL_MEM_FENCE);
			// Saving local data to global memory 
			oevent = async_work_group_copy(theArray+src_offset0, pDst0_out, local_size, 0);
			wait_group_events(1, &oevent);

		}
	}
	else	//first stage, sort inside one four
	{
		// 1 source & 1 DMA
		src_offset0 = wx * local_size;

		pSrc0_in = input0_local;
		pDst0_out = output0_local;		
		//barrier(CLK_LOCAL_MEM_FENCE);
		// Saving global data to local memory 
		ievent = async_work_group_copy(pSrc0_in, theArray+src_offset0, local_size, 0);
		wait_group_events(1, &ievent);

		func00(&(pSrc0_in[lx]), &(pDst0_out[lx]), i, dir);
		//func0(&(pSrc0_in[lx]), i, dir);
		//for(int i=0; i<local_size*2; i++)
		//	output0_local[i] = input0_local[i];

		//barrier(CLK_GLOBAL_MEM_FENCE);
		barrier(CLK_LOCAL_MEM_FENCE);
		// Saving local data to global memory 
		oevent = async_work_group_copy(theArray+src_offset0, pDst0_out, local_size, 0);
		wait_group_events(1, &oevent);
	}
}


__kernel 
void TEST_BARRIER_0(__global uint * theArray,
                    const uint stage, 
                    const uint passOfStage,
                    const uint direction)
{
    uint threadId = get_global_id(0);

	// local id 
	const int lx = get_local_id(0);

	// work group id
	const int wx = get_group_id(0);

	// local size
	const int local_size = get_local_size(0);

	__local uint input0_local[16];//BITONIC_LOCAL_SIZE_AMD*2];  
	__local uint input1_local[16];

	event_t ievent;
	event_t oevent;

	uint src_offset0, src_offset1;

	src_offset0 = 2* wx * local_size; 
	src_offset1 = src_offset0 + local_size; 

	uint leftId = 2*lx;
	uint rightId = leftId + 1;
	
	ievent = async_work_group_copy(input0_local, theArray+src_offset0, local_size*2, 0);
	wait_group_events(1, &ievent);	

	uint leftElement;
	uint rightElement; 
	uint greater;
	uint lesser;

	leftElement = input0_local[leftId];
	rightElement = input0_local[rightId];

	input0_local[leftId]  = rightElement;
	input0_local[rightId] = leftElement;

	barrier(CLK_LOCAL_MEM_FENCE);

	oevent = async_work_group_copy(theArray+src_offset0, input0_local, local_size*2, 0);
	wait_group_events(1, &oevent);	
}